"""
qm7 dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

QM7_MAT_UTL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7.mat"
QM7_CSV_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7.csv"
QM7B_MAT_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7b.mat"
GDB7_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb7.tar.gz"
QM7_TASKS = ["u0_atom"]


class _QM7Loader(_MolnetLoader):

  def create_dataset(self) -> Dataset:
    dataset_file = os.path.join(self.data_dir, "gdb7.sdf")
    if not os.path.exists(dataset_file):
      dc.utils.data_utils.download_url(url=GDB7_URL, dest_dir=self.data_dir)
      dc.utils.data_utils.untargz_file(
          os.path.join(self.data_dir, "gdb7.tar.gz"), self.data_dir)
    loader = dc.data.SDFLoader(
        tasks=self.tasks, featurizer=self.featurizer, sanitize=True)
    return loader.create_dataset(dataset_file, shard_size=8192)


def load_qm7(
    featurizer: Union[dc.feat.Featurizer, str] = dc.feat.CoulombMatrix(23),
    splitter: Union[dc.splits.Splitter, str, None] = 'random',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
  """Load QM7 dataset

  QM7 is a subset of GDB-13 (a database of nearly 1 billion
  stable and synthetically accessible organic molecules)
  containing up to 7 heavy atoms C, N, O, and S. The 3D
  Cartesian coordinates of the most stable conformations and
  their atomization energies were determined using ab-initio
  density functional theory (PBE0/tier2 basis set). This dataset
  also provided Coulomb matrices as calculated in [Rupp et al.
  PRL, 2012]:

  Stratified splitting is recommended for this dataset.

  The data file (.mat format, we recommend using `scipy.io.loadmat`
  for python users to load this original data) contains five arrays:

  - "X" - (7165 x 23 x 23), Coulomb matrices
  - "T" - (7165), atomization energies (unit: kcal/mol)
  - "P" - (5 x 1433), cross-validation splits as used in [Montavon et al.
    NIPS, 2012]
  - "Z" - (7165 x 23), atomic charges
  - "R" - (7165 x 23 x 3), cartesian coordinate (unit: Bohr) of each atom in
    the molecules

  Parameters
  ----------
  featurizer: Featurizer or str
    the featurizer to use for processing the data.  Alternatively you can pass
    one of the names from dc.molnet.featurizers as a shortcut.
  splitter: Splitter or str
    the splitter to use for splitting the data into training, validation, and
    test sets.  Alternatively you can pass one of the names from
    dc.molnet.splitters as a shortcut.  If this is None, all the data
    will be included in a single dataset.
  transformers: list of TransformerGenerators or strings
    the Transformers to apply to the data.  Each one is specified by a
    TransformerGenerator or, as a shortcut, one of the names from
    dc.molnet.transformers.
  reload: bool
    if True, the first call for a particular featurizer and splitter will cache
    the datasets to disk, and subsequent calls will reload the cached datasets.
  data_dir: str
    a directory to save the raw data in
  save_dir: str
    a directory to save the dataset in

  Note
  ----
  DeepChem 2.4.0 has turned on sanitization for this dataset by
  default.  For the QM7 dataset, this means that calling this
  function will return 6838 compounds instead of 7160 in the source
  dataset file.  This appears to be due to valence specification
  mismatches in the dataset that weren't caught in earlier more lax
  versions of RDKit.  Note that this may subtly affect benchmarking
  results on this
  dataset.

  References
  ----------
  .. [1] Rupp, Matthias, et al. "Fast and accurate modeling of molecular
     atomization energies with machine learning." Physical review letters
     108.5 (2012): 058301.
  .. [2] Montavon, Grégoire, et al. "Learning invariant representations of
     molecules for atomization energy prediction." Advances in Neural
     Information Proccessing Systems. 2012.
  """
  loader = _QM7Loader(featurizer, splitter, transformers, QM7_TASKS, data_dir,
                      save_dir, **kwargs)
  return loader.load_dataset('qm7', reload)



import numpy as np
import tensorflow as tf
from deepchem.data import NumpyDataset
from deepchem.feat import CircularFingerprint
from deepchem.models import KerasModel
from deepchem.models.losses import HingeLoss
from tensorflow.keras.layers import Input, Dense, Dropout, Activation, Lambda


class ScScoreModel(KerasModel):
  """
  https://pubs.acs.org/doi/abs/10.1021/acs.jcim.7b00622
  Several definitions of molecular complexity exist to facilitate prioritization
  of lead compounds, to identify diversity-inducing and complexifying reactions,
  and to guide retrosynthetic searches. In this work, we focus on synthetic
  complexity and reformalize its definition to correlate with the expected number
  of reaction steps required to produce a target molecule, with implicit knowledge
  about what compounds are reasonable starting materials. We train a neural
  network model on 12 million reactions from the Reaxys database to impose a
  pairwise inequality constraint enforcing the premise of this definition: that on
  average, the products of published chemical reactions should be more
  synthetically complex than their corresponding reactants. The learned metric
  (SCScore) exhibits highly desirable nonlinear behavior, particularly in
  recognizing increases in synthetic complexity throughout a number of linear
  synthetic routes.

  Our model here actually uses hingeloss instead of the shifted relu loss in
  https://github.com/connorcoley/scscore.

  This could cause issues differentiation issues with compounds that are "close"
  to each other in "complexity"

  """

  def __init__(self,
               n_features,
               layer_sizes=[300, 300, 300],
               dropouts=0.0,
               **kwargs):
    """
    Parameters
    ----------
    n_features: int
      number of features per molecule
    layer_sizes: list of int
      size of each hidden layer
    dropouts: int
      droupout to apply to each hidden layer
    kwargs
      This takes all kwards as TensorGraph
    """
    self.n_features = n_features
    self.layer_sizes = layer_sizes
    self.dropout = dropouts

    m1_features = Input(shape=(self.n_features,))
    m2_features = Input(shape=(self.n_features,))
    prev_layer1 = m1_features
    prev_layer2 = m2_features
    for layer_size in self.layer_sizes:
      layer = Dense(layer_size, activation=tf.nn.relu)
      prev_layer1 = layer(prev_layer1)
      prev_layer2 = layer(prev_layer2)
      if self.dropout > 0.0:
        prev_layer1 = Dropout(rate=self.dropout)(prev_layer1)
        prev_layer2 = Dropout(rate=self.dropout)(prev_layer2)

    readout_layer = Dense(1)
    readout_m1 = readout_layer(prev_layer1)
    readout_m2 = readout_layer(prev_layer2)
    outputs = [
        Lambda(lambda x: tf.sigmoid(x) * 4 + 1)(readout_m1),
        Lambda(lambda x: tf.sigmoid(x) * 4 + 1)(readout_m2),
        Lambda(lambda x: x[0] - x[1])([readout_m1, readout_m2])
    ]
    output_types = ['prediction', 'prediction', 'loss']
    model = tf.keras.Model(inputs=[m1_features, m2_features], outputs=outputs)
    super(ScScoreModel, self).__init__(
        model, HingeLoss(), output_types=output_types, **kwargs)

  def default_generator(self,
                        dataset,
                        epochs=1,
                        mode='fit',
                        deterministic=True,
                        pad_batches=True):
    for epoch in range(epochs):
      for (X_b, y_b, w_b, ids_b) in dataset.iterbatches(
          batch_size=self.batch_size,
          deterministic=deterministic,
          pad_batches=pad_batches):
        yield ([X_b[:, 0], X_b[:, 1]], [y_b], [w_b])

  def predict_mols(self, mols):
    featurizer = CircularFingerprint(
        size=self.n_features, radius=2, chiral=True)
    features = np.expand_dims(featurizer.featurize(mols), axis=1)
    features = np.concatenate([features, features], axis=1)
    ds = NumpyDataset(features, None, None, None)
    return self.predict(ds)[0][:, 0]



"""Utilities for handling PDBQT files."""

from typing import Dict, List, Optional, Set, Tuple
from deepchem.utils.typing import RDKitMol


def pdbqt_to_pdb(filename: Optional[str] = None,
                 pdbqt_data: Optional[List[str]] = None) -> str:
  """Extracts the PDB part of a pdbqt file as a string.

  Either `filename` or `pdbqt_data` must be provided. This function
  strips PDBQT charge information from the provided input.

  Parameters
  ----------
  filename: str, optional  (default None)
    Filename of PDBQT file
  pdbqt_data: List[str], optional (default None)
    Raw list of lines containing data from PDBQT file.

  Returns
  -------
  pdb_block: str
    String containing the PDB portion of pdbqt file.
  """
  if filename is not None and pdbqt_data is not None:
    raise ValueError("Only one of filename or pdbqt_data can be provided")
  elif filename is None and pdbqt_data is None:
    raise ValueError("Either filename or pdbqt_data must be provided")
  elif filename is not None:
    pdbqt_data = open(filename).readlines()

  pdb_block = ""
  # FIXME: Item "None" of "Optional[List[str]]" has no attribute "__iter__" (not iterable)
  for line in pdbqt_data:  # type: ignore
    pdb_block += "%s\n" % line[:66]
  return pdb_block


def convert_protein_to_pdbqt(mol: RDKitMol, outfile: str) -> None:
  """Convert a protein PDB file into a pdbqt file.

  Writes the extra PDBQT terms directly to `outfile`.

  Parameters
  ----------
  mol: rdkit.Chem.rdchem.Mol
    Protein molecule
  outfile: str
    filename which already has a valid pdb representation of mol
  """
  lines = [x.strip() for x in open(outfile).readlines()]
  out_lines = []
  for line in lines:
    if "ROOT" in line or "ENDROOT" in line or "TORSDOF" in line:
      out_lines.append("%s\n" % line)
      continue
    if not line.startswith("ATOM"):
      continue
    line = line[:66]
    atom_index = int(line[6:11])
    atom = mol.GetAtoms()[atom_index - 1]
    line = "%s    +0.000 %s\n" % (line, atom.GetSymbol().ljust(2))
    out_lines.append(line)
  with open(outfile, 'w') as fout:
    for line in out_lines:
      fout.write(line)


def _mol_to_graph(mol: RDKitMol):
  """Convert RDKit Mol to NetworkX graph

  Convert mol into a graph representation atoms are nodes, and bonds
  are vertices stored as graph

  Parameters
  ----------
  mol: rdkit.Chem.rdchem.Mol
    The molecule to convert into a graph.

  Returns
  -------
  graph: networkx.Graph
    Contains atoms indices as nodes, edges as bonds.

  Notes
  -----
  This function requires NetworkX to be installed.
  """
  try:
    import networkx as nx
  except ModuleNotFoundError:
    raise ImportError("This function requires NetworkX to be installed.")

  G = nx.Graph()
  num_atoms = mol.GetNumAtoms()
  G.add_nodes_from(range(num_atoms))
  for i in range(mol.GetNumBonds()):
    from_idx = mol.GetBonds()[i].GetBeginAtomIdx()
    to_idx = mol.GetBonds()[i].GetEndAtomIdx()
    G.add_edge(from_idx, to_idx)
  return G


def _get_rotatable_bonds(mol: RDKitMol) -> List[Tuple[int, int]]:
  """
  https://github.com/rdkit/rdkit/blob/f4529c910e546af590c56eba01f96e9015c269a6/Code/GraphMol/Descriptors/Lipinski.cpp#L107

  Taken from rdkit source to find which bonds are rotatable store
  rotatable bonds in (from_atom, to_atom)

  Parameters
  ----------
  mol: rdkit.Chem.rdchem.Mol
    Ligand molecule

  Returns
  -------
  rotatable_bonds: List[List[int, int]]
    List of rotatable bonds in molecule

  Notes
  -----
  This function requires RDKit to be installed.
  """
  try:
    from rdkit import Chem
    from rdkit.Chem import rdmolops
  except ModuleNotFoundError:
    raise ImportError("This function requires RDKit to be installed.")

  pattern = Chem.MolFromSmarts(
      "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])("
      "[CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])&!$([#7,O,S!D1]-!@[CD3]="
      "[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])&!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#"
      "*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])"
      "[CH3])]")
  rdmolops.FastFindRings(mol)
  rotatable_bonds = mol.GetSubstructMatches(pattern)
  return rotatable_bonds


def convert_mol_to_pdbqt(mol: RDKitMol, outfile: str) -> None:
  """Writes the provided ligand molecule to specified file in pdbqt format.

  Creates a torsion tree and write to pdbqt file. The torsion tree
  represents rotatable bonds in the molecule.

  Parameters
  ----------
  mol: rdkit.Chem.rdchem.Mol
    The molecule whose value is stored in pdb format in outfile
  outfile: str
    Filename for a valid pdb file with the extention .pdbqt

  Notes
  -----
  This function requires NetworkX to be installed.
  """
  try:
    import networkx as nx
  except ModuleNotFoundError:
    raise ImportError("This function requires NetworkX to be installed.")

  # Walk through the original file and extract ATOM/HETATM lines and
  # add PDBQT charge annotations.
  pdb_map = _create_pdb_map(outfile)
  graph = _mol_to_graph(mol)
  rotatable_bonds = _get_rotatable_bonds(mol)

  # Remove rotatable bonds from this molecule
  for bond in rotatable_bonds:
    graph.remove_edge(bond[0], bond[1])
  # Get the connected components now that the rotatable bonds have
  # been removed.
  components = [x for x in nx.connected_components(graph)]
  comp_map = _create_component_map(mol, components)

  used_partitions = set()
  lines = []
  # The root is the largest connected component.
  root = max(enumerate(components), key=lambda x: len(x[1]))[0]
  # Write the root component
  lines.append("ROOT\n")
  for atom in components[root]:
    lines.append(pdb_map[atom])
  lines.append("ENDROOT\n")
  # We've looked at the root, so take note of that
  used_partitions.add(root)
  for bond in rotatable_bonds:
    valid, next_partition = _valid_bond(used_partitions, bond, root, comp_map)
    if not valid:
      continue
    _dfs(used_partitions, next_partition, bond, components, rotatable_bonds,
         lines, pdb_map, comp_map)
  lines.append("TORSDOF %s" % len(rotatable_bonds))
  with open(outfile, 'w') as fout:
    for line in lines:
      fout.write(line)


def _create_pdb_map(outfile: str) -> Dict[int, str]:
  """Create a mapping from atom numbers to lines to write to pdbqt

  This is a map from rdkit atom number to its line in the pdb
  file. We also add the two additional columns required for
  pdbqt (charge, symbol).

  Note rdkit atoms are 0 indexed and pdb files are 1 indexed

  Parameters
  ----------
  outfile: str
    filename which already has a valid pdb representation of mol

  Returns
  -------
  pdb_map: Dict[int, str]
    Maps rdkit atom numbers to lines to be written to PDBQT file.
  """
  lines = [x.strip() for x in open(outfile).readlines()]
  lines = list(
      filter(lambda x: x.startswith("HETATM") or x.startswith("ATOM"), lines))
  lines = [x[:66] for x in lines]
  pdb_map = {}
  for line in lines:
    my_values = line.split()
    atom_number = int(my_values[1])
    atom_symbol = my_values[2]
    atom_symbol = ''.join([i for i in atom_symbol if not i.isdigit()])
    line = line.replace("HETATM", "ATOM  ")
    line = "%s    +0.000 %s\n" % (line, atom_symbol.ljust(2))
    pdb_map[atom_number - 1] = line
  return pdb_map


def _create_component_map(mol: RDKitMol,
                          components: List[List[int]]) -> Dict[int, int]:
  """Creates a map from atom ids to disconnected component id

  For each atom in `mol`, maps it to the id of the component in the
  molecule. The intent is that this is used on a molecule whose
  rotatable bonds have been removed. `components` is a list of the
  connected components after this surgery.

  Parameters
  ----------
  mol: rdkit.Chem.rdchem.Mol
    The molecule to find disconnected components in
  components: List[List[int]]
    List of connected components

  Returns
  -------
  comp_map: Dict[int, int]
    Maps atom ids to component ides
  """
  comp_map = {}
  for i in range(mol.GetNumAtoms()):
    for j in range(len(components)):
      if i in components[j]:
        comp_map[i] = j
        break
  return comp_map


def _dfs(used_partitions: Set[int], current_partition: int,
         bond: Tuple[int, int], components: List[List[int]],
         rotatable_bonds: List[Tuple[int, int]], lines: List[str],
         pdb_map: Dict[int, str], comp_map: Dict[int, int]) -> List[str]:
  """
  This function does a depth first search through the torsion tree

  Parameters
  ----------
  used_partions: Set[int]
    Partitions which have already been used
  current_partition: int
    The current partition to expand
  bond: Tuple[int, int]
    the bond which goes from the previous partition into this partition
  components: List[List[int]]
    List of connected components
  rotatable_bonds: List[Tuple[int, int]]
    List of rotatable bonds. This tuple is (from_atom, to_atom).
  lines: List[str]
    List of lines to write
  pdb_map: Dict[int, str]
    Maps atom numbers to PDBQT lines to write
  comp_map: Dict[int, int]
    Maps atom numbers to component numbers

  Returns
  -------
  lines: List[str]
    List of lines to write. This has more appended lines.
  """
  if comp_map[bond[1]] != current_partition:
    bond = (bond[1], bond[0])
  used_partitions.add(comp_map[bond[0]])
  used_partitions.add(comp_map[bond[1]])
  lines.append("BRANCH %4s %4s\n" % (bond[0] + 1, bond[1] + 1))
  for atom in components[current_partition]:
    lines.append(pdb_map[atom])
  for b in rotatable_bonds:
    valid, next_partition = \
      _valid_bond(used_partitions, b, current_partition, comp_map)
    if not valid:
      continue
    lines = _dfs(used_partitions, next_partition, b, components,
                 rotatable_bonds, lines, pdb_map, comp_map)
  lines.append("ENDBRANCH %4s %4s\n" % (bond[0] + 1, bond[1] + 1))
  return lines


def _valid_bond(used_partitions: Set[int], bond: Tuple[int, int],
                current_partition: int,
                comp_map: Dict[int, int]) -> Tuple[bool, int]:
  """Helper method to find next partition to explore.

  Used to check if a bond goes from the current partition into a
  partition that is not yet explored

  Parameters
  ----------
  used_partions: Set[int]
    Partitions which have already been used
  bond: Tuple[int, int]
    The bond to check if it goes to an unexplored partition.
    This tuple is (from_atom, to_atom).
  current_partition: int
    The current partition of the DFS
  comp_map: Dict[int, int]
    Maps atom ids to component ids

  Returns
  -------
  is_valid: bool
    Whether to exist the next partition or not
  next_partition: int
    The next partition to explore
  """
  part1 = comp_map[bond[0]]
  part2 = comp_map[bond[1]]
  if part1 != current_partition and part2 != current_partition:
    return False, 0
  if part1 == current_partition:
    next_partition = part2
  else:
    next_partition = part1
  return next_partition not in used_partitions, next_partition



"""Generative Adversarial Networks."""

from deepchem.models import KerasModel, layers, losses
from tensorflow.keras.layers import Input, Lambda, Layer, Softmax, Reshape, Multiply
from collections.abc import Sequence

import numpy as np
import tensorflow as tf
import time


class GAN(KerasModel):
  """Implements Generative Adversarial Networks.

  A Generative Adversarial Network (GAN) is a type of generative model.  It
  consists of two parts called the "generator" and the "discriminator".  The
  generator takes random noise as input and transforms it into an output that
  (hopefully) resembles the training data.  The discriminator takes a set of
  samples as input and tries to distinguish the real training samples from the
  ones created by the generator.  Both of them are trained together.  The
  discriminator tries to get better and better at telling real from false data,
  while the generator tries to get better and better at fooling the discriminator.

  In many cases there also are additional inputs to the generator and
  discriminator.  In that case it is known as a Conditional GAN (CGAN), since it
  learns a distribution that is conditional on the values of those inputs.  They
  are referred to as "conditional inputs".

  Many variations on this idea have been proposed, and new varieties of GANs are
  constantly being proposed.  This class tries to make it very easy to implement
  straightforward GANs of the most conventional types.  At the same time, it
  tries to be flexible enough that it can be used to implement many (but
  certainly not all) variations on the concept.

  To define a GAN, you must create a subclass that provides implementations of
  the following methods:

  get_noise_input_shape()
  get_data_input_shapes()
  create_generator()
  create_discriminator()

  If you want your GAN to have any conditional inputs you must also implement:

  get_conditional_input_shapes()

  The following methods have default implementations that are suitable for most
  conventional GANs.  You can override them if you want to customize their
  behavior:

  create_generator_loss()
  create_discriminator_loss()
  get_noise_batch()

  This class allows a GAN to have multiple generators and discriminators, a model
  known as MIX+GAN.  It is described in Arora et al., "Generalization and
  Equilibrium in Generative Adversarial Nets (GANs)" (https://arxiv.org/abs/1703.00573).
  This can lead to better models, and is especially useful for reducing mode
  collapse, since different generators can learn different parts of the
  distribution.  To use this technique, simply specify the number of generators
  and discriminators when calling the constructor.  You can then tell
  predict_gan_generator() which generator to use for predicting samples.
  """

  def __init__(self, n_generators=1, n_discriminators=1, **kwargs):
    """Construct a GAN.

    In addition to the parameters listed below, this class accepts all the
    keyword arguments from KerasModel.

    Parameters
    ----------
    n_generators: int
      the number of generators to include
    n_discriminators: int
      the number of discriminators to include
    """
    self.n_generators = n_generators
    self.n_discriminators = n_discriminators

    # Create the inputs.

    self.noise_input = Input(shape=self.get_noise_input_shape())
    self.data_input_layers = []
    for shape in self.get_data_input_shapes():
      self.data_input_layers.append(Input(shape=shape))
    self.data_inputs = [i.ref() for i in self.data_input_layers]
    self.conditional_input_layers = []
    for shape in self.get_conditional_input_shapes():
      self.conditional_input_layers.append(Input(shape=shape))
    self.conditional_inputs = [i.ref() for i in self.conditional_input_layers]

    # Create the generators.

    self.generators = []
    self.gen_variables = []
    generator_outputs = []
    for i in range(n_generators):
      generator = self.create_generator()
      self.generators.append(generator)
      generator_outputs.append(
          generator(
              _list_or_tensor([self.noise_input] +
                              self.conditional_input_layers)))
      self.gen_variables += generator.trainable_variables

    # Create the discriminators.

    self.discriminators = []
    self.discrim_variables = []
    discrim_train_outputs = []
    discrim_gen_outputs = []
    for i in range(n_discriminators):
      discriminator = self.create_discriminator()
      self.discriminators.append(discriminator)
      discrim_train_outputs.append(
          self._call_discriminator(discriminator, self.data_input_layers, True))
      for gen_output in generator_outputs:
        if tf.is_tensor(gen_output):
          gen_output = [gen_output]
        discrim_gen_outputs.append(
            self._call_discriminator(discriminator, gen_output, False))
      self.discrim_variables += discriminator.trainable_variables

    # Compute the loss functions.

    gen_losses = [self.create_generator_loss(d) for d in discrim_gen_outputs]
    discrim_losses = []
    for i in range(n_discriminators):
      for j in range(n_generators):
        discrim_losses.append(
            self.create_discriminator_loss(
                discrim_train_outputs[i],
                discrim_gen_outputs[i * n_generators + j]))
    if n_generators == 1 and n_discriminators == 1:
      total_gen_loss = gen_losses[0]
      total_discrim_loss = discrim_losses[0]
    else:
      # Create learnable weights for the generators and discriminators.

      gen_alpha = layers.Variable(np.ones((1, n_generators)), dtype=tf.float32)
      # We pass an input to the Variable layer to work around a bug in TF 1.14.
      gen_weights = Softmax()(gen_alpha([self.noise_input]))
      discrim_alpha = layers.Variable(np.ones((1, n_discriminators)),
                                      dtype=tf.float32)
      discrim_weights = Softmax()(discrim_alpha([self.noise_input]))

      # Compute the weighted errors

      weight_products = Reshape((n_generators * n_discriminators,))(Multiply()([
          Reshape((n_discriminators, 1))(discrim_weights),
          Reshape((1, n_generators))(gen_weights)
      ]))
      stacked_gen_loss = layers.Stack(axis=0)(gen_losses)
      stacked_discrim_loss = layers.Stack(axis=0)(discrim_losses)
      total_gen_loss = Lambda(lambda x: tf.reduce_sum(x[0] * x[1]))(
          [stacked_gen_loss, weight_products])
      total_discrim_loss = Lambda(lambda x: tf.reduce_sum(x[0] * x[1]))(
          [stacked_discrim_loss, weight_products])
      self.gen_variables += gen_alpha.trainable_variables
      self.discrim_variables += gen_alpha.trainable_variables
      self.discrim_variables += discrim_alpha.trainable_variables

      # Add an entropy term to the loss.

      entropy = Lambda(
          lambda x: -(tf.reduce_sum(tf.math.log(x[0])) / n_generators + tf.
                      reduce_sum(tf.math.log(x[1])) / n_discriminators))(
                          [gen_weights, discrim_weights])
      total_discrim_loss = Lambda(lambda x: x[0] + x[1])(
          [total_discrim_loss, entropy])

    # Create the Keras model.

    inputs = [self.noise_input
             ] + self.data_input_layers + self.conditional_input_layers
    outputs = [total_gen_loss, total_discrim_loss]
    self.gen_loss_fn = lambda outputs, labels, weights: outputs[0]
    self.discrim_loss_fn = lambda outputs, labels, weights: outputs[1]
    model = tf.keras.Model(inputs=inputs, outputs=outputs)
    super(GAN, self).__init__(model, self.gen_loss_fn, **kwargs)

  def _call_discriminator(self, discriminator, inputs, train):
    """Invoke the discriminator on a set of inputs.

    This is a separate method so WGAN can override it and also return the
    gradient penalty.
    """
    return discriminator(_list_or_tensor(inputs +
                                         self.conditional_input_layers))

  def get_noise_input_shape(self):
    """Get the shape of the generator's noise input layer.

    Subclasses must override this to return a tuple giving the shape of the
    noise input.  The actual Input layer will be created automatically.  The
    dimension corresponding to the batch size should be omitted.
    """
    raise NotImplementedError("Subclasses must implement this.")

  def get_data_input_shapes(self):
    """Get the shapes of the inputs for training data.

    Subclasses must override this to return a list of tuples, each giving the
    shape of one of the inputs.  The actual Input layers will be created
    automatically.  This list of shapes must also match the shapes of the
    generator's outputs.  The dimension corresponding to the batch size should
    be omitted.
    """
    raise NotImplementedError("Subclasses must implement this.")

  def get_conditional_input_shapes(self):
    """Get the shapes of any conditional inputs.

    Subclasses may override this to return a list of tuples, each giving the
    shape of one of the conditional inputs.  The actual Input layers will be
    created automatically.  The dimension corresponding to the batch size should
    be omitted.

    The default implementation returns an empty list, meaning there are no
    conditional inputs.
    """
    return []

  def get_noise_batch(self, batch_size):
    """Get a batch of random noise to pass to the generator.

    This should return a NumPy array whose shape matches the one returned by
    get_noise_input_shape().  The default implementation returns normally
    distributed values.  Subclasses can override this to implement a different
    distribution.
    """
    size = list(self.get_noise_input_shape())
    size = [batch_size] + size
    return np.random.normal(size=size)

  def create_generator(self):
    """Create and return a generator.

    Subclasses must override this to construct the generator.  The returned
    value should be a tf.keras.Model whose inputs are a batch of noise, followed
    by any conditional inputs.  The number and shapes of its outputs must match
    the return value from get_data_input_shapes(), since generated data must
    have the same form as training data.
    """
    raise NotImplementedError("Subclasses must implement this.")

  def create_discriminator(self):
    """Create and return a discriminator.

    Subclasses must override this to construct the discriminator.  The returned
    value should be a tf.keras.Model whose inputs are all data inputs, followed
    by any conditional inputs.  Its output should be a one dimensional tensor
    containing the probability of each sample being a training sample.
    """
    raise NotImplementedError("Subclasses must implement this.")

  def create_generator_loss(self, discrim_output):
    """Create the loss function for the generator.

    The default implementation is appropriate for most cases.  Subclasses can
    override this if the need to customize it.

    Parameters
    ----------
    discrim_output: Tensor
      the output from the discriminator on a batch of generated data.  This is
      its estimate of the probability that each sample is training data.

    Returns
    -------
    A Tensor equal to the loss function to use for optimizing the generator.
    """
    return Lambda(lambda x: -tf.reduce_mean(tf.math.log(x + 1e-10)))(
        discrim_output)

  def create_discriminator_loss(self, discrim_output_train, discrim_output_gen):
    """Create the loss function for the discriminator.

    The default implementation is appropriate for most cases.  Subclasses can
    override this if the need to customize it.

    Parameters
    ----------
    discrim_output_train: Tensor
      the output from the discriminator on a batch of training data.  This is
      its estimate of the probability that each sample is training data.
    discrim_output_gen: Tensor
      the output from the discriminator on a batch of generated data.  This is
      its estimate of the probability that each sample is training data.

    Returns
    -------
    A Tensor equal to the loss function to use for optimizing the discriminator.
    """
    return Lambda(lambda x: -tf.reduce_mean(
        tf.math.log(x[0] + 1e-10) + tf.math.log(1 - x[1] + 1e-10)))(
            [discrim_output_train, discrim_output_gen])

  def fit_gan(self,
              batches,
              generator_steps=1.0,
              max_checkpoints_to_keep=5,
              checkpoint_interval=1000,
              restore=False):
    """Train this model on data.

    Parameters
    ----------
    batches: iterable
      batches of data to train the discriminator on, each represented as a dict
      that maps Inputs to values.  It should specify values for all members of
      data_inputs and conditional_inputs.
    generator_steps: float
      the number of training steps to perform for the generator for each batch.
      This can be used to adjust the ratio of training steps for the generator
      and discriminator.  For example, 2.0 will perform two training steps for
      every batch, while 0.5 will only perform one training step for every two
      batches.
    max_checkpoints_to_keep: int
      the maximum number of checkpoints to keep.  Older checkpoints are discarded.
    checkpoint_interval: int
      the frequency at which to write checkpoints, measured in batches.  Set
      this to 0 to disable automatic checkpointing.
    restore: bool
      if True, restore the model from the most recent checkpoint before training
      it.
    """
    self._ensure_built()
    gen_train_fraction = 0.0
    discrim_error = 0.0
    gen_error = 0.0
    discrim_average_steps = 0
    gen_average_steps = 0
    time1 = time.time()
    if checkpoint_interval > 0:
      manager = tf.train.CheckpointManager(self._checkpoint, self.model_dir,
                                           max_checkpoints_to_keep)
    for feed_dict in batches:
      # Every call to fit_generator() will increment global_step, but we only
      # want it to get incremented once for the entire batch, so record the
      # value and keep resetting it.

      global_step = self.get_global_step()

      # Train the discriminator.

      inputs = [self.get_noise_batch(self.batch_size)]
      for input in self.data_input_layers:
        inputs.append(feed_dict[input.ref()])
      for input in self.conditional_input_layers:
        inputs.append(feed_dict[input.ref()])
      discrim_error += self.fit_generator([(inputs, [], [])],
                                          variables=self.discrim_variables,
                                          loss=self.discrim_loss_fn,
                                          checkpoint_interval=0,
                                          restore=restore)
      restore = False
      discrim_average_steps += 1

      # Train the generator.

      if generator_steps > 0.0:
        gen_train_fraction += generator_steps
        while gen_train_fraction >= 1.0:
          inputs = [self.get_noise_batch(self.batch_size)] + inputs[1:]
          gen_error += self.fit_generator([(inputs, [], [])],
                                          variables=self.gen_variables,
                                          checkpoint_interval=0)
          gen_average_steps += 1
          gen_train_fraction -= 1.0
      self._global_step.assign(global_step + 1)

      # Write checkpoints and report progress.

      if discrim_average_steps == checkpoint_interval:
        manager.save()
        discrim_loss = discrim_error / max(1, discrim_average_steps)
        gen_loss = gen_error / max(1, gen_average_steps)
        print(
            'Ending global_step %d: generator average loss %g, discriminator average loss %g'
            % (global_step, gen_loss, discrim_loss))
        discrim_error = 0.0
        gen_error = 0.0
        discrim_average_steps = 0
        gen_average_steps = 0

    # Write out final results.

    if checkpoint_interval > 0:
      if discrim_average_steps > 0 and gen_average_steps > 0:
        discrim_loss = discrim_error / discrim_average_steps
        gen_loss = gen_error / gen_average_steps
        print(
            'Ending global_step %d: generator average loss %g, discriminator average loss %g'
            % (global_step, gen_loss, discrim_loss))
      manager.save()
      time2 = time.time()
      print("TIMING: model fitting took %0.3f s" % (time2 - time1))

  def predict_gan_generator(self,
                            batch_size=1,
                            noise_input=None,
                            conditional_inputs=[],
                            generator_index=0):
    """Use the GAN to generate a batch of samples.

    Parameters
    ----------
    batch_size: int
      the number of samples to generate.  If either noise_input or
      conditional_inputs is specified, this argument is ignored since the batch
      size is then determined by the size of that argument.
    noise_input: array
      the value to use for the generator's noise input.  If None (the default),
      get_noise_batch() is called to generate a random input, so each call will
      produce a new set of samples.
    conditional_inputs: list of arrays
      the values to use for all conditional inputs.  This must be specified if
      the GAN has any conditional inputs.
    generator_index: int
      the index of the generator (between 0 and n_generators-1) to use for
      generating the samples.

    Returns
    -------
    An array (if the generator has only one output) or list of arrays (if it has
    multiple outputs) containing the generated samples.
    """
    if noise_input is not None:
      batch_size = len(noise_input)
    elif len(conditional_inputs) > 0:
      batch_size = len(conditional_inputs[0])
    if noise_input is None:
      noise_input = self.get_noise_batch(batch_size)
    inputs = [noise_input]
    inputs += conditional_inputs
    inputs = [i.astype(np.float32) for i in inputs]
    pred = self.generators[generator_index](_list_or_tensor(inputs),
                                            training=False)
    pred = pred.numpy()
    return pred


def _list_or_tensor(inputs):
  if len(inputs) == 1:
    return inputs[0]
  return inputs


class WGAN(GAN):
  """Implements Wasserstein Generative Adversarial Networks.

  This class implements Wasserstein Generative Adversarial Networks (WGANs) as
  described in Arjovsky et al., "Wasserstein GAN" (https://arxiv.org/abs/1701.07875).
  A WGAN is conceptually rather different from a conventional GAN, but in
  practical terms very similar.  It reinterprets the discriminator (often called
  the "critic" in this context) as learning an approximation to the Earth Mover
  distance between the training and generated distributions.  The generator is
  then trained to minimize that distance.  In practice, this just means using
  slightly different loss functions for training the generator and discriminator.

  WGANs have theoretical advantages over conventional GANs, and they often work
  better in practice.  In addition, the discriminator's loss function can be
  directly interpreted as a measure of the quality of the model.  That is an
  advantage over conventional GANs, where the loss does not directly convey
  information about the quality of the model.

  The theory WGANs are based on requires the discriminator's gradient to be
  bounded.  The original paper achieved this by clipping its weights.  This
  class instead does it by adding a penalty term to the discriminator's loss, as
  described in https://arxiv.org/abs/1704.00028.  This is sometimes found to
  produce better results.

  There are a few other practical differences between GANs and WGANs.  In a
  conventional GAN, the discriminator's output must be between 0 and 1 so it can
  be interpreted as a probability.  In a WGAN, it should produce an unbounded
  output that can be interpreted as a distance.

  When training a WGAN, you also should usually use a smaller value for
  generator_steps.  Conventional GANs rely on keeping the generator and
  discriminator "in balance" with each other.  If the discriminator ever gets
  too good, it becomes impossible for the generator to fool it and training
  stalls.  WGANs do not have this problem, and in fact the better the
  discriminator is, the easier it is for the generator to improve.  It therefore
  usually works best to perform several training steps on the discriminator for
  each training step on the generator.
  """

  def __init__(self, gradient_penalty=10.0, **kwargs):
    """Construct a WGAN.

    In addition to the following, this class accepts all the keyword arguments
    from GAN and KerasModel.

    Parameters
    ----------
    gradient_penalty: float
      the magnitude of the gradient penalty loss
    """
    self.gradient_penalty = gradient_penalty
    super(WGAN, self).__init__(**kwargs)

  def _call_discriminator(self, discriminator, inputs, train):
    if train:
      penalty = GradientPenaltyLayer(self, discriminator)
      return penalty(inputs, self.conditional_input_layers)
    return discriminator(_list_or_tensor(inputs +
                                         self.conditional_input_layers))

  def create_generator_loss(self, discrim_output):
    return Lambda(lambda x: tf.reduce_mean(x))(discrim_output)

  def create_discriminator_loss(self, discrim_output_train, discrim_output_gen):
    return Lambda(lambda x: tf.reduce_mean(x[0] - x[1]))(
        [discrim_output_train[0], discrim_output_gen]) + discrim_output_train[1]


class GradientPenaltyLayer(Layer):
  """Implements the gradient penalty loss term for WGANs."""

  def __init__(self, gan, discriminator, **kwargs):
    super(GradientPenaltyLayer, self).__init__(**kwargs)
    self.gan = gan
    self.discriminator = discriminator

  def call(self, inputs, conditional_inputs):
    with tf.GradientTape() as tape:
      for layer in inputs:
        tape.watch(layer)
      output = self.discriminator(_list_or_tensor(inputs + conditional_inputs))
    gradients = tape.gradient(output, inputs)
    gradients = [g for g in gradients if g is not None]
    if len(gradients) > 0:
      norm2 = 0.0
      for g in gradients:
        g2 = tf.square(g)
        dims = len(g.shape)
        if dims > 1:
          g2 = tf.reduce_sum(g2, axis=list(range(1, dims)))
        norm2 += g2
      penalty = tf.square(tf.sqrt(norm2) - 1.0)
      penalty = self.gan.gradient_penalty * tf.reduce_mean(penalty)
    else:
      penalty = 0.0
    return [output, penalty]



import numpy as np
import tensorflow as tf
from collections.abc import Sequence as SequenceCollection

import logging
import deepchem as dc
from deepchem.metrics import to_one_hot
from deepchem.models import KerasModel
from deepchem.models.layers import Stack
from deepchem.models.losses import SoftmaxCrossEntropy, L2Loss
from typing import Tuple, Iterable, List

logger = logging.getLogger(__name__)


class RobustMultitaskClassifier(KerasModel):
  """Implements a neural network for robust multitasking.

  The key idea of this model is to have bypass layers that feed
  directly from features to task output. This might provide some
  flexibility toroute around challenges in multitasking with
  destructive interference. 

  References
  ----------
  This technique was introduced in [1]_

  .. [1] Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

  """

  def __init__(self,
               n_tasks,
               n_features,
               layer_sizes=[1000],
               weight_init_stddevs=0.02,
               bias_init_consts=1.0,
               weight_decay_penalty=0.0,
               weight_decay_penalty_type="l2",
               dropouts=0.5,
               activation_fns=tf.nn.relu,
               n_classes=2,
               bypass_layer_sizes=[100],
               bypass_weight_init_stddevs=[.02],
               bypass_bias_init_consts=[1.],
               bypass_dropouts=[.5],
               **kwargs):
    """  Create a RobustMultitaskClassifier.

    Parameters
    ----------
    n_tasks: int
      number of tasks
    n_features: int
      number of features
    layer_sizes: list
      the size of each dense layer in the network.  The length of this list determines the number of layers.
    weight_init_stddevs: list or float
      the standard deviation of the distribution to use for weight initialization of each layer.  The length
      of this list should equal len(layer_sizes).  Alternatively this may be a single value instead of a list,
      in which case the same value is used for every layer.
    bias_init_consts: list or loat
      the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes).
      Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
    weight_decay_penalty: float
      the magnitude of the weight decay penalty to use
    weight_decay_penalty_type: str
      the type of penalty to use for weight decay, either 'l1' or 'l2'
    dropouts: list or float
      the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
      Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
    activation_fns: list or object
      the Tensorflow activation function to apply to each layer.  The length of this list should equal
      len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
      same value is used for every layer.
    n_classes: int
      the number of classes
    bypass_layer_sizes: list
      the size of each dense layer in the bypass network. The length of this list determines the number of bypass layers.
    bypass_weight_init_stddevs: list or float
      the standard deviation of the distribution to use for weight initialization of bypass layers.
      same requirements as weight_init_stddevs
    bypass_bias_init_consts: list or float
      the value to initialize the biases in bypass layers
      same requirements as bias_init_consts
    bypass_dropouts: list or float
      the dropout probablity to use for bypass layers.
      same requirements as dropouts
    """
    self.n_tasks = n_tasks
    self.n_features = n_features
    self.n_classes = n_classes
    n_layers = len(layer_sizes)
    if not isinstance(weight_init_stddevs, SequenceCollection):
      weight_init_stddevs = [weight_init_stddevs] * n_layers
    if not isinstance(bias_init_consts, SequenceCollection):
      bias_init_consts = [bias_init_consts] * n_layers
    if not isinstance(dropouts, SequenceCollection):
      dropouts = [dropouts] * n_layers
    if not isinstance(activation_fns, SequenceCollection):
      activation_fns = [activation_fns] * n_layers
    if weight_decay_penalty != 0.0:
      if weight_decay_penalty_type == 'l1':
        regularizer = tf.keras.regularizers.l1(weight_decay_penalty)
      else:
        regularizer = tf.keras.regularizers.l2(weight_decay_penalty)
    else:
      regularizer = None

    n_bypass_layers = len(bypass_layer_sizes)
    if not isinstance(bypass_weight_init_stddevs, SequenceCollection):
      bypass_weight_init_stddevs = [bypass_weight_init_stddevs
                                   ] * n_bypass_layers
    if not isinstance(bypass_bias_init_consts, SequenceCollection):
      bypass_bias_init_consts = [bypass_bias_init_consts] * n_bypass_layers
    if not isinstance(bypass_dropouts, SequenceCollection):
      bypass_dropouts = [bypass_dropouts] * n_bypass_layers
    bypass_activation_fns = [activation_fns[0]] * n_bypass_layers

    # Add the input features.
    mol_features = tf.keras.Input(shape=(n_features,))
    prev_layer = mol_features

    # Add the shared dense layers
    for size, weight_stddev, bias_const, dropout, activation_fn in zip(
        layer_sizes, weight_init_stddevs, bias_init_consts, dropouts,
        activation_fns):
      layer = tf.keras.layers.Dense(
          size,
          activation=activation_fn,
          kernel_initializer=tf.keras.initializers.TruncatedNormal(
              stddev=weight_stddev),
          bias_initializer=tf.constant_initializer(value=bias_const),
          kernel_regularizer=regularizer)(prev_layer)
      if dropout > 0.0:
        layer = tf.keras.layers.Dropout(rate=dropout)(layer)
      prev_layer = layer
    top_multitask_layer = prev_layer

    task_outputs = []
    for i in range(self.n_tasks):
      prev_layer = mol_features
      # Add task-specific bypass layers
      for size, weight_stddev, bias_const, dropout, activation_fn in zip(
          bypass_layer_sizes, bypass_weight_init_stddevs,
          bypass_bias_init_consts, bypass_dropouts, bypass_activation_fns):
        layer = tf.keras.layers.Dense(
            size,
            activation=activation_fn,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(
                stddev=weight_stddev),
            bias_initializer=tf.constant_initializer(value=bias_const),
            kernel_regularizer=regularizer)(prev_layer)
        if dropout > 0.0:
          layer = tf.keras.layers.Dropout(rate=dropout)(layer)
        prev_layer = layer
      top_bypass_layer = prev_layer

      if n_bypass_layers > 0:
        task_layer = tf.keras.layers.Concatenate(axis=1)(
            [top_multitask_layer, top_bypass_layer])
      else:
        task_layer = top_multitask_layer

      task_out = tf.keras.layers.Dense(n_classes)(task_layer)
      task_outputs.append(task_out)

    logits = Stack(axis=1)(task_outputs)
    output = tf.keras.layers.Softmax()(logits)
    model = tf.keras.Model(inputs=mol_features, outputs=[output, logits])
    super(RobustMultitaskClassifier,
          self).__init__(model,
                         SoftmaxCrossEntropy(),
                         output_types=['prediction', 'loss'],
                         **kwargs)

  def default_generator(self,
                        dataset,
                        epochs=1,
                        mode='fit',
                        deterministic=True,
                        pad_batches=True):
    for epoch in range(epochs):
      for (X_b, y_b, w_b,
           ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                         deterministic=deterministic,
                                         pad_batches=pad_batches):
        if y_b is not None:
          y_b = to_one_hot(y_b.flatten(),
                           self.n_classes).reshape(-1, self.n_tasks,
                                                   self.n_classes)
        yield ([X_b], [y_b], [w_b])

  def create_estimator_inputs(self, feature_columns, weight_column, features,
                              labels, mode):
    tensors = {}
    for layer, column in zip(self.features, feature_columns):
      tensors[layer] = tf.feature_column.input_layer(features, [column])
    if weight_column is not None:
      tensors[self.task_weights[0]] = tf.feature_column.input_layer(
          features, [weight_column])
    if labels is not None:
      tensors[self.labels[0]] = tf.one_hot(tf.cast(labels, tf.int32),
                                           self.n_classes)
    return tensors


class RobustMultitaskRegressor(KerasModel):
  """Implements a neural network for robust multitasking.

  The key idea of this model is to have bypass layers that feed
  directly from features to task output. This might provide some
  flexibility to route around challenges in multitasking with
  destructive interference.

  References
  ----------
  .. [1]   Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

  """

  def __init__(self,
               n_tasks,
               n_features,
               layer_sizes=[1000],
               weight_init_stddevs=0.02,
               bias_init_consts=1.0,
               weight_decay_penalty=0.0,
               weight_decay_penalty_type="l2",
               dropouts=0.5,
               activation_fns=tf.nn.relu,
               bypass_layer_sizes=[100],
               bypass_weight_init_stddevs=[.02],
               bypass_bias_init_consts=[1.],
               bypass_dropouts=[.5],
               **kwargs):
    """ Create a RobustMultitaskRegressor.

    Parameters
    ----------
    n_tasks: int
      number of tasks
    n_features: int
      number of features
    layer_sizes: list
      the size of each dense layer in the network.  The length of this list determines the number of layers.
    weight_init_stddevs: list or float
      the standard deviation of the distribution to use for weight initialization of each layer.  The length
      of this list should equal len(layer_sizes).  Alternatively this may be a single value instead of a list,
      in which case the same value is used for every layer.
    bias_init_consts: list or loat
      the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes).
      Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
    weight_decay_penalty: float
      the magnitude of the weight decay penalty to use
    weight_decay_penalty_type: str
      the type of penalty to use for weight decay, either 'l1' or 'l2'
    dropouts: list or float
      the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
      Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
    activation_fns: list or object
      the Tensorflow activation function to apply to each layer.  The length of this list should equal
      len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
      same value is used for every layer.
    bypass_layer_sizes: list
      the size of each dense layer in the bypass network. The length of this list determines the number of bypass layers.
    bypass_weight_init_stddevs: list or float
      the standard deviation of the distribution to use for weight initialization of bypass layers.
      same requirements as weight_init_stddevs
    bypass_bias_init_consts: list or float
      the value to initialize the biases in bypass layers
      same requirements as bias_init_consts
    bypass_dropouts: list or float
      the dropout probablity to use for bypass layers.
      same requirements as dropouts
    """
    self.n_tasks = n_tasks
    self.n_features = n_features
    n_layers = len(layer_sizes)
    if not isinstance(weight_init_stddevs, SequenceCollection):
      weight_init_stddevs = [weight_init_stddevs] * n_layers
    if not isinstance(bias_init_consts, SequenceCollection):
      bias_init_consts = [bias_init_consts] * n_layers
    if not isinstance(dropouts, SequenceCollection):
      dropouts = [dropouts] * n_layers
    if not isinstance(activation_fns, SequenceCollection):
      activation_fns = [activation_fns] * n_layers
    if weight_decay_penalty != 0.0:
      if weight_decay_penalty_type == 'l1':
        regularizer = tf.keras.regularizers.l1(weight_decay_penalty)
      else:
        regularizer = tf.keras.regularizers.l2(weight_decay_penalty)
    else:
      regularizer = None

    n_bypass_layers = len(bypass_layer_sizes)
    if not isinstance(bypass_weight_init_stddevs, SequenceCollection):
      bypass_weight_init_stddevs = [bypass_weight_init_stddevs
                                   ] * n_bypass_layers
    if not isinstance(bypass_bias_init_consts, SequenceCollection):
      bypass_bias_init_consts = [bypass_bias_init_consts] * n_bypass_layers
    if not isinstance(bypass_dropouts, SequenceCollection):
      bypass_dropouts = [bypass_dropouts] * n_bypass_layers
    bypass_activation_fns = [activation_fns[0]] * n_bypass_layers

    # Add the input features.
    mol_features = tf.keras.Input(shape=(n_features,))
    prev_layer = mol_features

    # Add the shared dense layers
    for size, weight_stddev, bias_const, dropout, activation_fn in zip(
        layer_sizes, weight_init_stddevs, bias_init_consts, dropouts,
        activation_fns):
      layer = tf.keras.layers.Dense(
          size,
          activation=activation_fn,
          kernel_initializer=tf.keras.initializers.TruncatedNormal(
              stddev=weight_stddev),
          bias_initializer=tf.constant_initializer(value=bias_const),
          kernel_regularizer=regularizer)(prev_layer)
      if dropout > 0.0:
        layer = tf.keras.layers.Dropout(rate=dropout)(layer)
      prev_layer = layer
    top_multitask_layer = prev_layer

    task_outputs = []
    for i in range(self.n_tasks):
      prev_layer = mol_features
      # Add task-specific bypass layers
      for size, weight_stddev, bias_const, dropout, activation_fn in zip(
          bypass_layer_sizes, bypass_weight_init_stddevs,
          bypass_bias_init_consts, bypass_dropouts, bypass_activation_fns):
        layer = tf.keras.layers.Dense(
            size,
            activation=activation_fn,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(
                stddev=weight_stddev),
            bias_initializer=tf.constant_initializer(value=bias_const),
            kernel_regularizer=regularizer)(prev_layer)
        if dropout > 0.0:
          layer = tf.keras.layers.Dropout(rate=dropout)(layer)
        prev_layer = layer
      top_bypass_layer = prev_layer

      if n_bypass_layers > 0:
        task_layer = tf.keras.layers.Concatenate(axis=1)(
            [top_multitask_layer, top_bypass_layer])
      else:
        task_layer = top_multitask_layer

      task_out = tf.keras.layers.Dense(1)(task_layer)
      task_outputs.append(task_out)

    outputs = Stack(axis=1)(task_outputs)
    model = tf.keras.Model(inputs=mol_features, outputs=outputs)
    super(RobustMultitaskRegressor, self).__init__(model,
                                                   L2Loss(),
                                                   output_types=['prediction'],
                                                   **kwargs)

  def default_generator(
      self,
      dataset: dc.data.Dataset,
      epochs: int = 1,
      mode: str = 'fit',
      deterministic: bool = True,
      pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
    for epoch in range(epochs):
      for (X_b, y_b, w_b,
           ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                         deterministic=deterministic,
                                         pad_batches=pad_batches):
        yield ([X_b], [y_b], [w_b])



"""
DGL-based MPNN for graph property prediction.
"""
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel


class MPNN(nn.Module):
  """Model for Graph Property Prediction.

  This model proceeds as follows:

  * Combine latest node representations and edge features in updating node representations,
    which involves multiple rounds of message passing
  * For each graph, compute its representation by combining the representations
    of all nodes in it, which involves a Set2Set layer.
  * Perform the final prediction using an MLP

  Examples
  --------

  >>> import deepchem as dc
  >>> import dgl
  >>> from deepchem.models.torch_models import MPNN
  >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
  >>> featurizer = dc.feat.MolGraphConvFeaturizer(use_edges=True)
  >>> graphs = featurizer.featurize(smiles)
  >>> print(type(graphs[0]))
  <class 'deepchem.feat.graph_data.GraphData'>
  >>> dgl_graphs = [graphs[i].to_dgl_graph(self_loop=True) for i in range(len(graphs))]
  >>> # Batch two graphs into a graph of two connected components
  >>> batch_dgl_graph = dgl.batch(dgl_graphs)
  >>> model = MPNN(n_tasks=1, mode='regression')
  >>> preds = model(batch_dgl_graph)
  >>> print(type(preds))
  <class 'torch.Tensor'>
  >>> preds.shape == (2, 1)
  True

  References
  ----------
  .. [1] Justin Gilmer, Samuel S. Schoenholz, Patrick F. Riley, Oriol Vinyals, George E. Dahl.
         "Neural Message Passing for Quantum Chemistry." ICML 2017.

  Notes
  -----
  This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
  (https://github.com/awslabs/dgl-lifesci) to be installed.
  """

  def __init__(self,
               n_tasks: int,
               node_out_feats: int = 64,
               edge_hidden_feats: int = 128,
               num_step_message_passing: int = 3,
               num_step_set2set: int = 6,
               num_layer_set2set: int = 3,
               mode: str = 'regression',
               number_atom_features: int = 30,
               number_bond_features: int = 11,
               n_classes: int = 2,
               nfeat_name: str = 'x',
               efeat_name: str = 'edge_attr'):
    """
    Parameters
    ----------
    n_tasks: int
      Number of tasks.
    node_out_feats: int
      The length of the final node representation vectors. Default to 64.
    edge_hidden_feats: int
      The length of the hidden edge representation vectors. Default to 128.
    num_step_message_passing: int
      The number of rounds of message passing. Default to 3.
    num_step_set2set: int
      The number of set2set steps. Default to 6.
    num_layer_set2set: int
      The number of set2set layers. Default to 3.
    mode: str
      The model type, 'classification' or 'regression'. Default to 'regression'.
    number_atom_features: int
      The length of the initial atom feature vectors. Default to 30.
    number_bond_features: int
      The length of the initial bond feature vectors. Default to 11.
    n_classes: int
      The number of classes to predict per task
      (only used when ``mode`` is 'classification'). Default to 2.
    nfeat_name: str
      For an input graph ``g``, the model assumes that it stores node features in
      ``g.ndata[nfeat_name]`` and will retrieve input node features from that.
      Default to 'x'.
    efeat_name: str
      For an input graph ``g``, the model assumes that it stores edge features in
      ``g.edata[efeat_name]`` and will retrieve input edge features from that.
      Default to 'edge_attr'.
    """
    try:
      import dgl
    except:
      raise ImportError('This class requires dgl.')
    try:
      import dgllife
    except:
      raise ImportError('This class requires dgllife.')

    if mode not in ['classification', 'regression']:
      raise ValueError("mode must be either 'classification' or 'regression'")

    super(MPNN, self).__init__()

    self.n_tasks = n_tasks
    self.mode = mode
    self.n_classes = n_classes
    self.nfeat_name = nfeat_name
    self.efeat_name = efeat_name
    if mode == 'classification':
      out_size = n_tasks * n_classes
    else:
      out_size = n_tasks

    from dgllife.model import MPNNPredictor as DGLMPNNPredictor

    self.model = DGLMPNNPredictor(
        node_in_feats=number_atom_features,
        edge_in_feats=number_bond_features,
        node_out_feats=node_out_feats,
        edge_hidden_feats=edge_hidden_feats,
        n_tasks=out_size,
        num_step_message_passing=num_step_message_passing,
        num_step_set2set=num_step_set2set,
        num_layer_set2set=num_layer_set2set)

  def forward(self, g):
    """Predict graph labels

    Parameters
    ----------
    g: DGLGraph
      A DGLGraph for a batch of graphs. It stores the node features in
      ``dgl_graph.ndata[self.nfeat_name]`` and edge features in
      ``dgl_graph.edata[self.efeat_name]``.

    Returns
    -------
    torch.Tensor
      The model output.

      * When self.mode = 'regression',
        its shape will be ``(dgl_graph.batch_size, self.n_tasks)``.
      * When self.mode = 'classification', the output consists of probabilities
        for classes. Its shape will be
        ``(dgl_graph.batch_size, self.n_tasks, self.n_classes)`` if self.n_tasks > 1;
        its shape will be ``(dgl_graph.batch_size, self.n_classes)`` if self.n_tasks is 1.
    torch.Tensor, optional
      This is only returned when self.mode = 'classification', the output consists of the
      logits for classes before softmax.
    """
    node_feats = g.ndata[self.nfeat_name]
    edge_feats = g.edata[self.efeat_name]
    out = self.model(g, node_feats, edge_feats)

    if self.mode == 'classification':
      if self.n_tasks == 1:
        logits = out.view(-1, self.n_classes)
        softmax_dim = 1
      else:
        logits = out.view(-1, self.n_tasks, self.n_classes)
        softmax_dim = 2
      proba = F.softmax(logits, dim=softmax_dim)
      return proba, logits
    else:
      return out


class MPNNModel(TorchModel):
  """Model for graph property prediction

  This model proceeds as follows:

  * Combine latest node representations and edge features in updating node representations,
    which involves multiple rounds of message passing
  * For each graph, compute its representation by combining the representations
    of all nodes in it, which involves a Set2Set layer.
  * Perform the final prediction using an MLP

  Examples
  --------
  >>> import deepchem as dc
  >>> from deepchem.models.torch_models import MPNNModel
  >>> # preparing dataset
  >>> smiles = ["C1CCC1", "CCC"]
  >>> labels = [0., 1.]
  >>> featurizer = dc.feat.MolGraphConvFeaturizer(use_edges=True)
  >>> X = featurizer.featurize(smiles)
  >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
  >>> # training model
  >>> model = MPNNModel(mode='classification', n_tasks=1,
  ...                  batch_size=16, learning_rate=0.001)
  >>> loss =  model.fit(dataset, nb_epoch=5)

  References
  ----------
  .. [1] Justin Gilmer, Samuel S. Schoenholz, Patrick F. Riley, Oriol Vinyals, George E. Dahl.
         "Neural Message Passing for Quantum Chemistry." ICML 2017.

  Notes
  -----
  This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
  (https://github.com/awslabs/dgl-lifesci) to be installed.
  """

  def __init__(self,
               n_tasks: int,
               node_out_feats: int = 64,
               edge_hidden_feats: int = 128,
               num_step_message_passing: int = 3,
               num_step_set2set: int = 6,
               num_layer_set2set: int = 3,
               mode: str = 'regression',
               number_atom_features: int = 30,
               number_bond_features: int = 11,
               n_classes: int = 2,
               self_loop: bool = False,
               **kwargs):
    """
    Parameters
    ----------
    n_tasks: int
      Number of tasks.
    node_out_feats: int
      The length of the final node representation vectors. Default to 64.
    edge_hidden_feats: int
      The length of the hidden edge representation vectors. Default to 128.
    num_step_message_passing: int
      The number of rounds of message passing. Default to 3.
    num_step_set2set: int
      The number of set2set steps. Default to 6.
    num_layer_set2set: int
      The number of set2set layers. Default to 3.
    mode: str
      The model type, 'classification' or 'regression'. Default to 'regression'.
    number_atom_features: int
      The length of the initial atom feature vectors. Default to 30.
    number_bond_features: int
      The length of the initial bond feature vectors. Default to 11.
    n_classes: int
      The number of classes to predict per task
      (only used when ``mode`` is 'classification'). Default to 2.
    self_loop: bool
      Whether to add self loops for the nodes, i.e. edges from nodes to themselves.
      Generally, an MPNNModel does not require self loops. Default to False.
    kwargs
      This can include any keyword argument of TorchModel.
    """
    model = MPNN(
        n_tasks=n_tasks,
        node_out_feats=node_out_feats,
        edge_hidden_feats=edge_hidden_feats,
        num_step_message_passing=num_step_message_passing,
        num_step_set2set=num_step_set2set,
        num_layer_set2set=num_layer_set2set,
        mode=mode,
        number_atom_features=number_atom_features,
        number_bond_features=number_bond_features,
        n_classes=n_classes)
    if mode == 'regression':
      loss: Loss = L2Loss()
      output_types = ['prediction']
    else:
      loss = SparseSoftmaxCrossEntropy()
      output_types = ['prediction', 'loss']
    super(MPNNModel, self).__init__(
        model, loss=loss, output_types=output_types, **kwargs)

    self._self_loop = self_loop

  def _prepare_batch(self, batch):
    """Create batch data for MPNN.

    Parameters
    ----------
    batch: tuple
      The tuple is ``(inputs, labels, weights)``.

    Returns
    -------
    inputs: DGLGraph
      DGLGraph for a batch of graphs.
    labels: list of torch.Tensor or None
      The graph labels.
    weights: list of torch.Tensor or None
      The weights for each sample or sample/task pair converted to torch.Tensor.
    """
    try:
      import dgl
    except:
      raise ImportError('This class requires dgl.')

    inputs, labels, weights = batch
    dgl_graphs = [
        graph.to_dgl_graph(self_loop=self._self_loop) for graph in inputs[0]
    ]
    inputs = dgl.batch(dgl_graphs).to(self.device)
    _, labels, weights = super(MPNNModel, self)._prepare_batch(([], labels,
                                                                weights))
    return inputs, labels, weights



import time
import numpy as np
import tensorflow as tf
from collections.abc import Sequence as SequenceCollection

import logging
from deepchem.metrics import to_one_hot
from deepchem.metrics import from_one_hot
from deepchem.models import KerasModel, layers
from deepchem.models.losses import L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.keras_model import _StandardLoss
from tensorflow.keras.layers import Input, Dense, Dropout, ReLU, Concatenate, Add, Multiply, Softmax

logger = logging.getLogger(__name__)


class ProgressiveMultitaskRegressor(KerasModel):
  """Implements a progressive multitask neural network for regression.

  Progressive networks allow for multitask learning where each task
  gets a new column of weights. As a result, there is no exponential
  forgetting where previous tasks are ignored.

  References
  ----------
  See [1]_ for a full description of the progressive architecture

  .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
         arXiv:1606.04671 (2016).
  """

  def __init__(self,
               n_tasks,
               n_features,
               alpha_init_stddevs=0.02,
               layer_sizes=[1000],
               weight_init_stddevs=0.02,
               bias_init_consts=1.0,
               weight_decay_penalty=0.0,
               weight_decay_penalty_type="l2",
               dropouts=0.5,
               activation_fns=tf.nn.relu,
               n_outputs=1,
               **kwargs):
    """Creates a progressive network.

    Only listing parameters specific to progressive networks here.

    Parameters
    ----------
    n_tasks: int
      Number of tasks
    n_features: int
      Number of input features
    alpha_init_stddevs: list
      List of standard-deviations for alpha in adapter layers.
    layer_sizes: list
      the size of each dense layer in the network.  The length of this list determines the number of layers.
    weight_init_stddevs: list or float
      the standard deviation of the distribution to use for weight initialization of each layer.  The length
      of this list should equal len(layer_sizes)+1.  The final element corresponds to the output layer.
      Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
    bias_init_consts: list or float
      the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes)+1.
      The final element corresponds to the output layer.  Alternatively this may be a single value instead of a list,
      in which case the same value is used for every layer.
    weight_decay_penalty: float
      the magnitude of the weight decay penalty to use
    weight_decay_penalty_type: str
      the type of penalty to use for weight decay, either 'l1' or 'l2'
    dropouts: list or float
      the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
      Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
    activation_fns: list or object
      the Tensorflow activation function to apply to each layer.  The length of this list should equal
      len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
      same value is used for every layer.
    """

    if weight_decay_penalty != 0.0:
      raise ValueError('Weight decay is not currently supported')
    self.n_tasks = n_tasks
    self.n_features = n_features
    self.layer_sizes = layer_sizes
    self.alpha_init_stddevs = alpha_init_stddevs
    self.weight_init_stddevs = weight_init_stddevs
    self.bias_init_consts = bias_init_consts
    self.dropouts = dropouts
    self.activation_fns = activation_fns
    self.n_outputs = n_outputs

    n_layers = len(layer_sizes)
    if not isinstance(weight_init_stddevs, SequenceCollection):
      self.weight_init_stddevs = [weight_init_stddevs] * n_layers
    if not isinstance(alpha_init_stddevs, SequenceCollection):
      self.alpha_init_stddevs = [alpha_init_stddevs] * n_layers
    if not isinstance(bias_init_consts, SequenceCollection):
      self.bias_init_consts = [bias_init_consts] * n_layers
    if not isinstance(dropouts, SequenceCollection):
      self.dropouts = [dropouts] * n_layers
    if not isinstance(activation_fns, SequenceCollection):
      self.activation_fns = [activation_fns] * n_layers

    # Add the input features.
    mol_features = Input(shape=(n_features,))

    all_layers = {}
    outputs = []
    self._task_layers = []
    for task in range(self.n_tasks):
      task_layers = []
      for i in range(n_layers):
        if i == 0:
          prev_layer = mol_features
        else:
          prev_layer = all_layers[(i - 1, task)]
          if task > 0:
            lateral_contrib, trainables = self.add_adapter(all_layers, task, i)
            task_layers.extend(trainables)

        dense = Dense(layer_sizes[i],
                      kernel_initializer=tf.keras.initializers.TruncatedNormal(
                          stddev=self.weight_init_stddevs[i]),
                      bias_initializer=tf.constant_initializer(
                          value=self.bias_init_consts[i]))
        layer = dense(prev_layer)
        task_layers.append(dense)

        if i > 0 and task > 0:
          layer = Add()([layer, lateral_contrib])
        assert self.activation_fns[i] is tf.nn.relu, "Only ReLU is supported"
        layer = ReLU()(layer)
        if self.dropouts[i] > 0.0:
          layer = Dropout(self.dropouts[i])(layer)
        all_layers[(i, task)] = layer

      prev_layer = all_layers[(n_layers - 1, task)]
      dense = Dense(n_outputs,
                    kernel_initializer=tf.keras.initializers.TruncatedNormal(
                        stddev=self.weight_init_stddevs[-1]),
                    bias_initializer=tf.constant_initializer(
                        value=self.bias_init_consts[-1]))
      layer = dense(prev_layer)
      task_layers.append(dense)

      if task > 0:
        lateral_contrib, trainables = self.add_adapter(all_layers, task,
                                                       n_layers)
        task_layers.extend(trainables)
        layer = Add()([layer, lateral_contrib])
      output_layer = self.create_output(layer)
      outputs.append(output_layer)
      self._task_layers.append(task_layers)

    outputs = layers.Stack(axis=1)(outputs)
    model = tf.keras.Model(inputs=mol_features, outputs=outputs)
    super(ProgressiveMultitaskRegressor,
          self).__init__(model, self.create_loss(), **kwargs)

  def create_loss(self):
    return L2Loss()

  def create_output(self, layer):
    return layer

  def add_adapter(self, all_layers, task, layer_num):
    """Add an adapter connection for given task/layer combo"""
    i = layer_num
    prev_layers = []
    trainable_layers = []
    # Handle output layer
    if i < len(self.layer_sizes):
      layer_sizes = self.layer_sizes
      alpha_init_stddev = self.alpha_init_stddevs[i]
      weight_init_stddev = self.weight_init_stddevs[i]
      bias_init_const = self.bias_init_consts[i]
    elif i == len(self.layer_sizes):
      layer_sizes = self.layer_sizes + [self.n_outputs]
      alpha_init_stddev = self.alpha_init_stddevs[-1]
      weight_init_stddev = self.weight_init_stddevs[-1]
      bias_init_const = self.bias_init_consts[-1]
    else:
      raise ValueError("layer_num too large for add_adapter.")
    # Iterate over all previous tasks.
    for prev_task in range(task):
      prev_layers.append(all_layers[(i - 1, prev_task)])
    # prev_layers is a list with elements of size
    # (batch_size, layer_sizes[i-1])
    if len(prev_layers) == 1:
      prev_layer = prev_layers[0]
    else:
      prev_layer = Concatenate(axis=1)(prev_layers)
    alpha = layers.Variable(
        tf.random.truncated_normal((1,), stddev=alpha_init_stddev))
    trainable_layers.append(alpha)

    prev_layer = Multiply()([prev_layer, alpha([prev_layer])])
    dense1 = Dense(
        layer_sizes[i - 1],
        kernel_initializer=tf.keras.initializers.TruncatedNormal(
            stddev=weight_init_stddev),
        bias_initializer=tf.constant_initializer(value=bias_init_const))
    prev_layer = dense1(prev_layer)
    trainable_layers.append(dense1)

    dense2 = Dense(layer_sizes[i],
                   kernel_initializer=tf.keras.initializers.TruncatedNormal(
                       stddev=weight_init_stddev),
                   use_bias=False)
    prev_layer = dense2(prev_layer)
    trainable_layers.append(dense2)

    return prev_layer, trainable_layers

  def fit(self,
          dataset,
          nb_epoch=10,
          max_checkpoints_to_keep=5,
          checkpoint_interval=1000,
          deterministic=False,
          restore=False,
          **kwargs):
    for task in range(self.n_tasks):
      self.fit_task(dataset,
                    task,
                    nb_epoch=nb_epoch,
                    max_checkpoints_to_keep=max_checkpoints_to_keep,
                    checkpoint_interval=checkpoint_interval,
                    deterministic=deterministic,
                    restore=restore,
                    **kwargs)

  def fit_task(self,
               dataset,
               task,
               nb_epoch=10,
               max_checkpoints_to_keep=5,
               checkpoint_interval=1000,
               deterministic=False,
               restore=False,
               **kwargs):
    """Fit one task."""
    shape = dataset.get_shape()
    batch = [[np.zeros((self.batch_size,) + s[1:])] for s in shape]
    self._create_training_ops(batch)
    generator = self.default_generator(dataset,
                                       epochs=nb_epoch,
                                       deterministic=deterministic)
    variables = []
    for layer in self._task_layers[task]:
      variables += layer.trainable_variables
    loss = TaskLoss(self.model, self.create_loss(), task)
    self.fit_generator(generator,
                       max_checkpoints_to_keep,
                       checkpoint_interval,
                       restore,
                       variables=variables,
                       loss=loss)


class ProgressiveMultitaskClassifier(ProgressiveMultitaskRegressor):
  """Implements a progressive multitask neural network for classification.

  Progressive Networks: https://arxiv.org/pdf/1606.04671v3.pdf

  Progressive networks allow for multitask learning where each task
  gets a new column of weights. As a result, there is no exponential
  forgetting where previous tasks are ignored.

  """

  def __init__(self,
               n_tasks,
               n_features,
               alpha_init_stddevs=0.02,
               layer_sizes=[1000],
               weight_init_stddevs=0.02,
               bias_init_consts=1.0,
               weight_decay_penalty=0.0,
               weight_decay_penalty_type="l2",
               dropouts=0.5,
               activation_fns=tf.nn.relu,
               **kwargs):
    n_outputs = 2
    super(ProgressiveMultitaskClassifier,
          self).__init__(n_tasks,
                         n_features,
                         alpha_init_stddevs=alpha_init_stddevs,
                         layer_sizes=layer_sizes,
                         weight_init_stddevs=weight_init_stddevs,
                         bias_init_consts=bias_init_consts,
                         weight_decay_penalty=weight_decay_penalty,
                         weight_decay_penalty_type=weight_decay_penalty_type,
                         dropouts=dropouts,
                         activation_fns=activation_fns,
                         n_outputs=n_outputs,
                         **kwargs)

  def create_loss(self):
    return SparseSoftmaxCrossEntropy()

  def create_output(self, layer):
    return Softmax()(layer)


class TaskLoss(_StandardLoss):

  def __init__(self, model, loss, task):
    super(TaskLoss, self).__init__(model, loss)
    self.task = task

  def __call__(self, outputs, labels, weights):
    outputs = [t[:, self.task] for t in outputs]
    labels = [t[:, self.task] for t in labels]
    weights = [t[:, self.task] for t in weights]
    return super(TaskLoss, self).__call__(outputs, labels, weights)
